from typing import Optional

from akimbo.utils import NoAttributes

try:
    import cudf
    import cupy
except ImportError:
    cudf = NoAttributes()
    cupy = NoAttributes()
import awkward as ak
import numpy
import pyarrow
from awkward._backends.cupy import CupyBackend
from awkward._backends.numpy import NumpyBackend

# COPIED from awkward/studies/cudf-to-awkward.py

######################### stripped-down copy of src/awkward/_connect/pyarrow.py


_string_like = (
    pyarrow.string(),
    pyarrow.large_string(),
    pyarrow.binary(),
    pyarrow.large_binary(),
)

_pyarrow_to_numpy_dtype = {
    pyarrow.date32(): (True, numpy.dtype("M8[D]")),
    pyarrow.date64(): (False, numpy.dtype("M8[ms]")),
    pyarrow.time32("s"): (True, numpy.dtype("M8[s]")),
    pyarrow.time32("ms"): (True, numpy.dtype("M8[ms]")),
    pyarrow.time64("us"): (False, numpy.dtype("M8[us]")),
    pyarrow.time64("ns"): (False, numpy.dtype("M8[ns]")),
    pyarrow.timestamp("s"): (False, numpy.dtype("M8[s]")),
    pyarrow.timestamp("ms"): (False, numpy.dtype("M8[ms]")),
    pyarrow.timestamp("us"): (False, numpy.dtype("M8[us]")),
    pyarrow.timestamp("ns"): (False, numpy.dtype("M8[ns]")),
    pyarrow.duration("s"): (False, numpy.dtype("m8[s]")),
    pyarrow.duration("ms"): (False, numpy.dtype("m8[ms]")),
    pyarrow.duration("us"): (False, numpy.dtype("m8[us]")),
    pyarrow.duration("ns"): (False, numpy.dtype("m8[ns]")),
}


def revertable(modified, original):
    modified.__pyarrow_original = original
    return modified


def remove_optiontype(akarray):
    return akarray.__pyarrow_original


def popbuffers_finalize(out, array, validbits, generate_bitmasks, fix_offsets=True):
    # Every buffer from Arrow must be offsets-corrected.
    if fix_offsets and (array.offset != 0 or len(array) != len(out)):
        out = out[array.offset : array.offset + len(array)]

    # Everything must leave popbuffers as option-type; the mask_node will be
    # removed by the next level up in popbuffers recursion if appropriate.

    if validbits is None and generate_bitmasks:
        # ceildiv(len(out), 8) = -(len(out) // -8)
        validbits = numpy.full(-(len(out) // -8), numpy.uint8(0xFF), dtype=numpy.uint8)

    if validbits is None:
        return revertable(ak.contents.UnmaskedArray.simplified(out), out)
    else:
        return revertable(
            ak.contents.BitMaskedArray.simplified(
                ak.index.IndexU8(numpy.frombuffer(validbits, dtype=numpy.uint8)),
                out,
                valid_when=True,
                length=len(out),
                lsb_order=True,
            ),
            out,
        )


def popbuffers(paarray, arrow_type, buffers, generate_bitmasks):
    ### Beginning of the big if-elif-elif chain!

    if isinstance(arrow_type, pyarrow.lib.DictionaryType):
        masked_index = popbuffers(
            paarray.indices,
            arrow_type.index_type,
            buffers,
            generate_bitmasks,
        )
        index = masked_index.content.data

        if not isinstance(masked_index, ak.contents.UnmaskedArray):
            mask = masked_index.mask_as_bool(valid_when=False)
            if mask.any():
                index = numpy.asarray(index, copy=True)
                index[mask] = -1

        content = handle_arrow(paarray.dictionary, generate_bitmasks)

        parameters = {"__array__": "categorical"}

        return revertable(
            ak.contents.IndexedOptionArray.simplified(
                ak.index.Index(index),
                content,
                parameters=parameters,
            ),
            ak.contents.IndexedArray(
                ak.index.Index(index),
                remove_optiontype(content) if content.is_option else content,
                parameters=parameters,
            ),
        )

    elif isinstance(arrow_type, pyarrow.lib.FixedSizeListType):
        assert arrow_type.num_buffers == 1
        validbits = buffers.pop(0)

        akcontent = popbuffers(
            paarray.values, arrow_type.value_type, buffers, generate_bitmasks
        )

        if not arrow_type.value_field.nullable:
            # strip the dummy option-type node
            akcontent = remove_optiontype(akcontent)

        out = ak.contents.RegularArray(akcontent, arrow_type.list_size, parameters=None)
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    elif isinstance(arrow_type, (pyarrow.lib.LargeListType, pyarrow.lib.ListType)):
        assert arrow_type.num_buffers == 2
        validbits = buffers.pop(0)
        paoffsets = buffers.pop(0)

        if isinstance(arrow_type, pyarrow.lib.LargeListType):
            akoffsets = ak.index.Index64(numpy.frombuffer(paoffsets, dtype=numpy.int64))
        else:
            akoffsets = ak.index.Index32(numpy.frombuffer(paoffsets, dtype=numpy.int32))

        akcontent = popbuffers(
            paarray.values, arrow_type.value_type, buffers, generate_bitmasks
        )

        if not arrow_type.value_field.nullable:
            # strip the dummy option-type node
            akcontent = remove_optiontype(akcontent)

        out = ak.contents.ListOffsetArray(akoffsets, akcontent, parameters=None)
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.MapType):
        # FIXME: make a ListOffsetArray of 2-tuples with __array__ == "sorted_map".
        # (Make sure the keys are sorted).
        raise NotImplementedError

    elif isinstance(
        arrow_type, (pyarrow.lib.Decimal128Type, pyarrow.lib.Decimal256Type)
    ):
        # Note: Decimal128Type and Decimal256Type are subtypes of FixedSizeBinaryType.
        # NumPy doesn't support decimal: https://github.com/numpy/numpy/issues/9789
        raise ValueError(
            "Arrow arrays containing pyarrow.decimal128 or pyarrow.decimal256 types can't be converted into Awkward Arrays"
        )

    elif isinstance(arrow_type, pyarrow.lib.FixedSizeBinaryType):
        assert arrow_type.num_buffers == 2
        validbits = buffers.pop(0)
        pacontent = buffers.pop(0)

        parameters = {"__array__": "bytestring"}
        sub_parameters = {"__array__": "byte"}

        out = ak.contents.RegularArray(
            ak.contents.NumpyArray(
                numpy.frombuffer(pacontent, dtype=numpy.uint8),
                parameters=sub_parameters,
                backend=NumpyBackend.instance(),
            ),
            arrow_type.byte_width,
            parameters=parameters,
        )
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    elif arrow_type in _string_like:
        assert arrow_type.num_buffers == 3
        validbits = buffers.pop(0)
        paoffsets = buffers.pop(0)
        pacontent = buffers.pop(0)

        if arrow_type in _string_like[::2]:
            akoffsets = ak.index.Index32(numpy.frombuffer(paoffsets, dtype=numpy.int32))
        else:
            akoffsets = ak.index.Index64(numpy.frombuffer(paoffsets, dtype=numpy.int64))

        if arrow_type in _string_like[:2]:
            parameters = {"__array__": "string"}
            sub_parameters = {"__array__": "char"}
        else:
            parameters = {"__array__": "bytestring"}
            sub_parameters = {"__array__": "byte"}

        out = ak.contents.ListOffsetArray(
            akoffsets,
            ak.contents.NumpyArray(
                numpy.frombuffer(pacontent, dtype=numpy.uint8),
                parameters=sub_parameters,
                backend=NumpyBackend.instance(),
            ),
            parameters=parameters,
        )
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.StructType):
        assert arrow_type.num_buffers == 1
        validbits = buffers.pop(0)

        keys = []
        contents = []
        for i in range(arrow_type.num_fields):
            field = arrow_type[i]
            field_name = field.name
            keys.append(field_name)

            akcontent = popbuffers(
                paarray.field(field_name), field.type, buffers, generate_bitmasks
            )
            if not field.nullable:
                # strip the dummy option-type node
                akcontent = remove_optiontype(akcontent)
            contents.append(akcontent)

        out = ak.contents.RecordArray(
            contents, keys, length=len(paarray), parameters=None
        )
        return popbuffers_finalize(
            out, paarray, validbits, generate_bitmasks, fix_offsets=False
        )

    elif isinstance(arrow_type, pyarrow.lib.UnionType):
        if isinstance(arrow_type, pyarrow.lib.SparseUnionType):
            assert arrow_type.num_buffers == 2
            validbits = buffers.pop(0)
            nptags = numpy.frombuffer(buffers.pop(0), dtype=numpy.int8)
            npindex = numpy.arange(len(nptags), dtype=numpy.int32)
        else:
            assert arrow_type.num_buffers == 3
            validbits = buffers.pop(0)
            nptags = numpy.frombuffer(buffers.pop(0), dtype=numpy.int8)
            npindex = numpy.frombuffer(buffers.pop(0), dtype=numpy.int32)

        akcontents = []
        for i in range(arrow_type.num_fields):
            field = arrow_type[i]
            akcontent = popbuffers(
                paarray.field(i), field.type, buffers, generate_bitmasks
            )

            if not field.nullable:
                # strip the dummy option-type node
                akcontent = remove_optiontype(akcontent)
            akcontents.append(akcontent)

        out = ak.contents.UnionArray.simplified(
            ak.index.Index8(nptags),
            ak.index.Index32(npindex),
            akcontents,
            parameters=None,
        )
        return popbuffers_finalize(out, paarray, None, generate_bitmasks)

    elif arrow_type == pyarrow.null():
        validbits = buffers.pop(0)
        assert arrow_type.num_fields == 0

        # This is already an option-type and offsets-corrected, so no popbuffers_finalize.
        return ak.contents.IndexedOptionArray(
            ak.index.Index64(numpy.full(len(paarray), -1, dtype=numpy.int64)),
            ak.contents.EmptyArray(parameters=None),
            parameters=None,
        )

    elif arrow_type == pyarrow.bool_():
        assert arrow_type.num_buffers == 2
        validbits = buffers.pop(0)
        bitdata = buffers.pop(0)

        bytedata = numpy.unpackbits(
            numpy.frombuffer(bitdata, dtype=numpy.uint8), bitorder="little"
        )

        out = ak.contents.NumpyArray(
            bytedata.view(numpy.bool_),
            parameters=None,
            backend=NumpyBackend.instance(),
        )
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.DataType):
        assert arrow_type.num_buffers == 2
        validbits = buffers.pop(0)
        data = buffers.pop(0)

        to64, dt = _pyarrow_to_numpy_dtype.get(str(arrow_type), (False, None))
        if to64:
            data = numpy.astype(
                numpy.frombuffer(data, dtype=numpy.int32), dtype=numpy.int64
            )
        if dt is None:
            dt = arrow_type.to_pandas_dtype()

        out = ak.contents.NumpyArray(
            numpy.frombuffer(data, dtype=dt),
            parameters=None,
            backend=NumpyBackend.instance(),
        )
        return popbuffers_finalize(out, paarray, validbits, generate_bitmasks)

    else:
        raise TypeError(f"unrecognized Arrow array type: {arrow_type!r}")


def handle_arrow(obj, generate_bitmasks):
    buffers = obj.buffers()
    out = popbuffers(obj, obj.type, buffers, generate_bitmasks)
    assert len(buffers) == 0
    return out


def pyarrow_to_awkward(
    pyarrow_array: pyarrow.lib.Array,
    generate_bitmasks=False,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    ctx = ak._layout.HighLevelContext(behavior=behavior, attrs=attrs).finalize()

    out = handle_arrow(pyarrow_array, generate_bitmasks)
    if isinstance(out, ak.contents.UnmaskedArray):
        out = remove_optiontype(out)

    def remove_revertable(layout, **kwargs):
        if hasattr(layout, "__pyarrow_original"):
            del layout.__pyarrow_original

    ak._do.recursively_apply(out, remove_revertable)

    return ctx.wrap(out, highlevel=highlevel)


######################### equivalent for CuDF


def recurse_finalize(
    out: ak.contents.Content,
    column: cudf.core.column.column.ColumnBase,
    validbits: Optional[cudf.core.buffer.buffer.Buffer],
    generate_bitmasks: bool,
    fix_offsets: bool = True,
):
    # Every buffer from Arrow must be offsets-corrected.
    if fix_offsets and (column.offset != 0 or len(column) != len(out)):
        out = out[column.offset : column.offset + len(column)]

    if validbits is None:
        return revertable(ak.contents.UnmaskedArray.simplified(out), out)
    else:
        return revertable(
            ak.contents.BitMaskedArray.simplified(
                ak.index.IndexU8(cupy.asarray(validbits)),
                out,
                valid_when=True,
                length=len(out),
                lsb_order=True,
            ),
            out,
        )


def recurse(
    column: cudf.core.column.column.ColumnBase,
    arrow_type: pyarrow.lib.DataType,
    generate_bitmasks: bool,
):
    if isinstance(column, cudf.core.column.CategoricalColumn):
        validbits = column.base_mask

        paindex = column.base_children[-1]
        masked_index = recurse(paindex, arrow_type_of(paindex), generate_bitmasks)
        index = masked_index.content.data

        if not isinstance(masked_index, ak.contents.UnmaskedArray):
            mask = masked_index.mask_as_bool(valid_when=False)
            if mask.any():
                index = cupy.asarray(index, copy=True)
                index[mask] = -1

        pacats = column.categories
        content = recurse(pacats, arrow_type_of(pacats), generate_bitmasks)

        if index.dtype == cupy.dtype(cupy.int64):
            akindex1 = ak.index.Index64(index)
            akindex2 = akindex1
        elif index.dtype == cupy.dtype(cupy.uint32):
            akindex1 = ak.index.Index64(index.astype(cupy.int64))
            akindex2 = ak.index.IndexU32(index)
        elif index.dtype == cupy.dtype(cupy.int32):
            akindex1 = ak.index.Index32(index)
            akindex2 = akindex1
        else:
            akindex1 = ak.index.Index64(index.astype(cupy.int64))
            akindex2 = akindex1

        return revertable(
            ak.contents.IndexedOptionArray.simplified(
                akindex1,
                content,
                parameters={"__array__": "categorical"},
            ),
            ak.contents.IndexedArray(
                akindex2,
                remove_optiontype(content) if content.is_option else content,
                parameters={"__array__": "categorical"},
            ),
        )

    elif isinstance(arrow_type, pyarrow.lib.FixedSizeListType):
        validbits = column.base_mask

        akcontent = recurse(
            column.base_children[-1], arrow_type.value_type, generate_bitmasks
        )

        if not arrow_type.value_field.nullable:
            # strip the dummy option-type node
            akcontent = remove_optiontype(akcontent)

        out = ak.contents.RegularArray(akcontent, arrow_type.list_size, parameters=None)
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif isinstance(arrow_type, (pyarrow.lib.LargeListType, pyarrow.lib.ListType)):
        validbits = column.base_mask
        paoffsets = column.offsets.base_data

        if isinstance(arrow_type, pyarrow.lib.LargeListType):
            akoffsets = ak.index.Index64(cupy.asarray(paoffsets).view(cupy.int64))
        else:
            akoffsets = ak.index.Index32(cupy.asarray(paoffsets).view(cupy.int32))

        akcontent = recurse(
            column.base_children[-1], arrow_type.value_type, generate_bitmasks
        )

        if not arrow_type.value_field.nullable:
            # strip the dummy option-type node
            akcontent = remove_optiontype(akcontent)

        out = ak.contents.ListOffsetArray(akoffsets, akcontent, parameters=None)
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.MapType):
        # FIXME: make a ListOffsetArray of 2-tuples with __array__ == "sorted_map".
        # (Make sure the keys are sorted).
        raise NotImplementedError

    elif isinstance(
        arrow_type, (pyarrow.lib.Decimal128Type, pyarrow.lib.Decimal256Type)
    ):
        # Note: Decimal128Type and Decimal256Type are subtypes of FixedSizeBinaryType.
        # NumPy doesn't support decimal: https://github.com/numpy/numpy/issues/9789
        raise ValueError(
            "Arrow arrays containing pyarrow.decimal128 or pyarrow.decimal256 types can't be converted into Awkward Arrays"
        )

    elif isinstance(arrow_type, pyarrow.lib.FixedSizeBinaryType):
        validbits = column.base_mask
        pacontent = column.base_data

        parameters = {"__array__": "bytestring"}
        sub_parameters = {"__array__": "byte"}

        out = ak.contents.RegularArray(
            ak.contents.NumpyArray(
                cupy.asarray(pacontent),
                parameters=sub_parameters,
                backend=CupyBackend.instance(),
            ),
            arrow_type.byte_width,
            parameters=parameters,
        )
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif arrow_type in _string_like:
        validbits = column.base_mask

        paoffsets = column.base_children[-1]
        pacontent = column.base_data

        if arrow_type in _string_like[::2]:
            akoffsets = ak.index.Index32(cupy.asarray(paoffsets).view(cupy.int32))
        else:
            akoffsets = ak.index.Index64(cupy.asarray(paoffsets).view(cupy.int64))

        if arrow_type in _string_like[:2]:
            parameters = {"__array__": "string"}
            sub_parameters = {"__array__": "char"}
        else:
            parameters = {"__array__": "bytestring"}
            sub_parameters = {"__array__": "byte"}

        out = ak.contents.ListOffsetArray(
            akoffsets,
            ak.contents.NumpyArray(
                cupy.asarray(pacontent),
                parameters=sub_parameters,
                backend=CupyBackend.instance(),
            ),
            parameters=parameters,
        )
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.StructType):
        validbits = column.base_mask

        keys = []
        contents = []
        for i in range(arrow_type.num_fields):
            field = arrow_type[i]
            field_name = field.name
            keys.append(field_name)

            akcontent = recurse(column.base_children[i], field.type, generate_bitmasks)
            if not field.nullable:
                # strip the dummy option-type node
                akcontent = remove_optiontype(akcontent)
            contents.append(akcontent)

        out = ak.contents.RecordArray(
            contents, keys, length=len(column), parameters=None
        )
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.UnionType):
        raise NotImplementedError

    elif arrow_type == pyarrow.null():
        validbits = column.base_mask

        # This is already an option-type and offsets-corrected, so no popbuffers_finalize.
        return ak.contents.IndexedOptionArray(
            ak.index.Index64(cupy.full(len(column), -1, dtype=cupy.int64)),
            ak.contents.EmptyArray(parameters=None),
            parameters=None,
        )

    elif arrow_type == pyarrow.bool_():
        validbits = column.base_mask

        ## boolean data from CuDF differs from Arrow: it's represented as bytes, not bits!
        # bitdata = column.base_data
        # bytedata = cupy.unpackbits(cupy.asarray(bitdata), bitorder="little")
        bytedata = cupy.asarray(column.base_data)

        out = ak.contents.NumpyArray(
            cupy.asarray(bytedata).view(cupy.bool_),
            parameters=None,
            backend=CupyBackend.instance(),
        )
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    elif isinstance(arrow_type, pyarrow.lib.DataType):
        validbits = column.base_mask

        to64, dt = _pyarrow_to_numpy_dtype.get(str(arrow_type), (False, None))
        if dt is None:
            dt = arrow_type.to_pandas_dtype()
        if to64:
            data = cupy.asarray(column.base_data).view(cupy.int32).astype(cupy.int64)
        else:
            data = cupy.asarray(column.base_data)

        out = ak.contents.NumpyArray(
            data.view(dt),
            parameters=None,
            backend=CupyBackend.instance(),
        )
        return recurse_finalize(out, column, validbits, generate_bitmasks)

    else:
        raise TypeError(f"unrecognized Arrow array type: {arrow_type!r}")


def arrow_type_of(column):
    dtype = column.dtype

    if isinstance(column, cudf.core.column.StringColumn):
        return pyarrow.string()

    elif isinstance(column, cudf.core.column.CategoricalColumn):
        return None  # deal with it in `recurse` for nesting-generality

    elif isinstance(dtype, numpy.dtype):
        if dtype == numpy.dtype(object):
            raise TypeError("Python object type encountered in CuDF Series")
        else:
            return pyarrow.from_numpy_dtype(dtype)

    else:
        return dtype.to_arrow()


def handle_cudf(cudf_series: cudf.core.series.Series, generate_bitmasks):
    column = cudf_series._data[cudf_series.name]
    return recurse(column, arrow_type_of(column), generate_bitmasks)


def cudf_to_awkward(
    cudf_series: cudf.core.series.Series,
    generate_bitmasks=False,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    ctx = ak._layout.HighLevelContext(behavior=behavior, attrs=attrs).finalize()

    out = handle_cudf(cudf_series, generate_bitmasks)
    if isinstance(out, ak.contents.UnmaskedArray):
        out = remove_optiontype(out)

    def remove_revertable(layout, **kwargs):
        if hasattr(layout, "__pyarrow_original"):
            del layout.__pyarrow_original

    ak._do.recursively_apply(out, remove_revertable)

    return ctx.wrap(out, highlevel=highlevel)


######################### testing


if __name__ == "__main__":
    # tests numerics, lists, records, and option-type, but not union-type
    examples = [
        [False, True, True],  # booleans are special (1-bit)
        [1.1, 2.2, 3.3],
        [[False, True, True], [], [True, False]],
        [[1, 2, 3], [], [4, 5]],
        [[[1, 2], [3]], [], [[]], [[4], [], [5, 6, 7]], [[8, 9]]],
        [{"x": 1}, {"x": 2}, {"x": 3}],
        [{"x": 1.1, "y": []}, {"x": 2.2, "y": [1]}, {"x": 3.3, "y": [1, 2]}],
        [[{"x": 1}, {"x": 2}, {"x": 3}], [], [{"x": 4}, {"x": 5}]],
        ["This", "is", "a", "string", "array", ".", ""],
        [["This", "is", "a"], ["nested"], ["string", "array", ".", ""]],
        [None, None, None, None, None],
        [False, True, None, True],
        [1.1, 2.2, None, 3.3],
        [[False, True, None, True], [], [True, False]],
        [[False, True, True], None, [], [True, False]],
        [[1, 2, None, 3], [], [4, 5]],
        [[1, 2, 3], None, [], [4, 5]],
        [[[1, 2, None], [3]], [], [[]], [[4], [], [5, 6, 7]], [[8, 9]]],
        [[[1, 2], None, [3]], [], [[]], [[4], [], [5, 6, 7]], [[8, 9]]],
        [[[1, 2], [3]], None, [], [[]], [[4], [], [5, 6, 7]], [[8, 9]]],
        [{"x": 1}, {"x": None}, {"x": 3}],
        [{"x": 1}, {"x": 2}, None, {"x": 3}],
        [{"x": 1.1, "y": []}, {"x": None, "y": [1]}, {"x": 3.3, "y": [1, 2]}],
        [{"x": 1.1, "y": []}, {"x": 2.2, "y": [1, None]}, {"x": 3.3, "y": [1, 2]}],
        [{"x": 1.1, "y": []}, {"x": 2.2, "y": [1]}, None, {"x": 3.3, "y": [1, 2]}],
        [[{"x": 1}, {"x": None}, {"x": 3}], [], [{"x": 4}, {"x": 5}]],
        [[{"x": 1}, {"x": 2}, None, {"x": 3}], [], [{"x": 4}, {"x": 5}]],
        [[{"x": 1}, {"x": 2}, {"x": 3}], None, [], [{"x": 4}, {"x": 5}]],
        ["This", "is", "a", None, "string", "array", ".", ""],
        [["This", "is", "a", None], ["nested"], ["string", "array", ".", ""]],
        [["This", "is", "a"], None, ["nested"], ["string", "array", ".", ""]],
        numpy.array(["2024-01-01", "2024-01-02"], dtype="datetime64[s]"),
        numpy.array([1, 2, 3], dtype="timedelta64[s]"),
    ]

    for example in examples:
        print(f"---- {example}")
        df = cudf.DataFrame({"column": example})

        awkward_array = cudf_to_awkward(df["column"])
        assert ak.backend(awkward_array) == "cuda"
        assert awkward_array.tolist() == list(example), awkward_array.show(type=True)



from __future__ import annotations

import functools
import operator
from typing import Callable

import awkward as ak
import numpy as np
import pyarrow.compute as pc

from akimbo.apply_tree import dec, match_any, numeric, run_with_transform
from akimbo.ops import join, rename
from akimbo.utils import to_ak_layout

methods = [
    _ for _ in (dir(ak)) if not _.startswith(("_", "ak_")) and not _[0].isupper()
]


def radd(left, right):
    return right + left


def rsub(left, right):
    return right - left


def rmul(left, right):
    return right * left


def rdiv(left, right):
    return right / left


def rtruediv(left, right):
    return right / left


def rfloordiv(left, right):
    return right // left


def rmod(left, right):
    # check if right is a string as % is the string
    # formatting operation; this is a TypeError
    # otherwise perform the op
    if isinstance(right, str):
        typ = type(left).__name__
        raise TypeError(f"{typ} cannot perform the operation mod")

    return right % left


def rdivmod(left, right):
    return divmod(right, left)


def rpow(left, right):
    return right**left


def rand_(left, right):
    return operator.and_(right, left)


def ror_(left, right):
    return operator.or_(right, left)


def rxor(left, right):
    return operator.xor(right, left)


class AbstractMethodError(NotImplementedError):
    pass


class ArithmeticMixin:
    @classmethod
    def _create_op(cls, op):
        raise AbstractMethodError(cls)

    @classmethod
    def _add_arithmetic_ops(cls) -> None:
        setattr(cls, "__add__", cls._create_op(operator.add))
        setattr(cls, "__radd__", cls._create_op(radd))
        setattr(cls, "__sub__", cls._create_op(operator.sub))
        setattr(cls, "__rsub__", cls._create_op(rsub))
        setattr(cls, "__mul__", cls._create_op(operator.mul))
        setattr(cls, "__rmul__", cls._create_op(rmul))
        setattr(cls, "__pow__", cls._create_op(operator.pow))
        setattr(cls, "__rpow__", cls._create_op(rpow))
        setattr(cls, "__mod__", cls._create_op(operator.mod))
        setattr(cls, "__rmod__", cls._create_op(rmod))
        setattr(cls, "__floordiv__", cls._create_op(operator.floordiv))
        setattr(cls, "__rfloordiv__", cls._create_op(rfloordiv))
        setattr(cls, "__truediv__", cls._create_op(operator.truediv))
        setattr(cls, "__rtruediv__", cls._create_op(rtruediv))
        setattr(cls, "__divmod__", cls._create_op(divmod))
        setattr(cls, "__rdivmod__", cls._create_op(rdivmod))

    @classmethod
    def _add_comparison_ops(cls) -> None:
        setattr(cls, "__eq__", cls._create_op(operator.eq))
        setattr(cls, "__ne__", cls._create_op(operator.ne))
        setattr(cls, "__lt__", cls._create_op(operator.lt))
        setattr(cls, "__gt__", cls._create_op(operator.gt))
        setattr(cls, "__le__", cls._create_op(operator.le))
        setattr(cls, "__ge__", cls._create_op(operator.ge))

    @classmethod
    def _add_logical_ops(cls) -> None:
        setattr(cls, "__and__", cls._create_op(operator.and_))
        setattr(cls, "__rand__", cls._create_op(rand_))
        setattr(cls, "__or__", cls._create_op(operator.or_))
        setattr(cls, "__ror__", cls._create_op(ror_))
        setattr(cls, "__xor__", cls._create_op(operator.xor))
        setattr(cls, "__rxor__", cls._create_op(rxor))

    @classmethod
    def _add_all(cls):
        cls._add_logical_ops()
        cls._add_arithmetic_ops()
        cls._add_comparison_ops()


class EagerAccessor(ArithmeticMixin):
    """Bring the awkward API to dataframes and series"""

    series_type = ()
    dataframe_type = ()
    subaccessors = {}
    funcs = {"cast": dec(pc.cast, inmode="arrow"), "join": join, "rename": rename}

    def __init__(self, obj, behavior=None, subaccessor=None):
        self.subaccessor = subaccessor
        self._obj = obj
        self._behavior = behavior

    def __call__(self, *args, behavior=None, **kwargs):
        return EagerAccessor(self._obj, behavior=behavior)

    def unexplode(self, *cols: tuple[str, ...], outname="grouped") -> ak.Array:
        """Repack "exploded" form dataframes into lists of structs

        This is the inverse of the regular dataframe explode() process.

        Uses arrow directly.
        """
        from akimbo.utils import rec_list_swap

        pa_arr = self.arrow
        # TODO: this does not work on cuDF as here we use arrow directly
        # TODO: pandas indexes are pre-grouped cat-like structures
        cols = list(cols)
        if set(cols) - set(pa_arr.column_names):
            raise ValueError(
                "One or more grouping column (%s) not in available columns %s",
                cols,
                pa_arr.column_names,
            )
        outcols = [(_, "list") for _ in pa_arr.column_names if _ not in cols]
        if not outcols:
            raise ValueError("Cannot group on all available columns")
        outcols2 = [f"{_[0]}_list" for _ in outcols]
        grouped = pa_arr.group_by(cols).aggregate(outcols)
        akarr = ak.from_arrow(grouped)
        akarr2 = akarr[outcols2]
        akarr2.layout._fields = [_[0] for _ in outcols]
        struct = rec_list_swap(akarr2)
        final = ak.with_field(akarr[cols], struct, outname)

        return self.to_output(final).ak.unpack()

    @classmethod
    def _create_op(cls, op):
        def run(self, *args, **kwargs):
            if self.subaccessor:
                # defer
                op2 = op(self.subaccessors[self.subaccessor](), None)
                return self.__getattr__(op2)(*args, **kwargs)
            args = [
                to_ak_layout(_) if isinstance(_, (str, int, float, np.number)) else _
                for _ in args
            ]
            return self.transform(op, *args, match=numeric, **kwargs)

        return run

    def __getitem__(self, item):
        return self.__getattr__(operator.getitem)(item)

    def __array_ufunc__(self, *args, where=None, out=None, **kwargs):
        # includes operator overload like df.ak + 1
        ufunc, call, inputs, *callargs = args
        if out is not None or call != "__call__":
            raise NotImplementedError

        return self.__getattr__(ufunc)(*callargs, where=where, **kwargs)

    def __dir__(self) -> list[str]:
        if self.subaccessor is not None:
            return dir(self.subaccessors[self.subaccessor]())
        cls_meth = [_ for _ in dir(self.dataframe_type) if not _.startswith("_")]
        this_meth = [_ for _ in dir(type(self)) if not _.startswith("_")]
        return sorted(methods + cls_meth + this_meth + list(self.funcs))

    def transform(
        self,
        fn: callable,
        *others,
        where=None,
        match=match_any,
        inmode="array",
        **kwargs,
    ):
        def f(arr, *others, **kwargs):
            return run_with_transform(
                arr, fn, match=match, others=others, inmode=inmode, **kwargs
            )

        return self.__getattr__(f)(*others, where=where, **kwargs)

    def apply(self, fn: Callable, *others, where=None, **kwargs):
        return self.__getattr__(fn)(*others, where=where, **kwargs)

    def with_behavior(self, behavior, where=()):
        """Assign a behavior to this array-of-records"""
        # TODO: compare usage with sub-accessors
        # TODO: implement where= (assign directly to ._paraneters["__record__"]
        #  of output's layout. In this case, behaviour is a dict of locations to apply to.
        #  and we can continually add to it (or accept a dict)
        return type(self)(self._obj, behavior)

    with_name = with_behavior  # alias - this is the upstream name

    asa = with_behavior  # favoured name

    def __array_function__(self, *args, **kwargs):
        return self.array.__array_function__(*args, **kwargs)

    @classmethod
    def register_accessor(cls, name: str, klass: type):
        # TODO: check clobber?
        cls.subaccessors[name] = klass

    def _pre(self, item):
        arr = self.array
        if callable(item):
            func = item
        elif item in self.funcs:
            func = self.funcs[item]
        elif hasattr(arr, item) and callable(getattr(arr, item)):
            func = getattr(type(arr), item)
        elif self.subaccessor:
            func = getattr(self.subaccessors[self.subaccessor](), item)
        elif hasattr(ak, item):
            func = getattr(ak, item)
        else:
            raise AttributeError(item)
        return func, (arr,)

    @classmethod
    def _to_arr(cls, other):
        # make eager ak arrays for any standard input
        if isinstance(other, (cls.series_type, cls.dataframe_type)):
            return other.ak.array
        if isinstance(other, cls):
            return other.array
        return other

    def __getattr__(self, item, frame=False):
        if not self.subaccessor and item in self.subaccessors:
            return type(self)(self._obj, subaccessor=item)
        func, args = self._pre(item)
        frame = isinstance(self._obj, self.dataframe_type)

        @functools.wraps(func)
        def f(*others, where=None, **kwargs):
            others = [self._to_arr(other) for other in others]
            kwargs = {k: self._to_arr(v) for k, v in kwargs.items()}
            if where:
                arr0 = args[0]
                ar0 = [_[where] for _ in args]
            else:
                arr0 = None
                ar0 = args
            ak_arr = func(*ar0, *others, **kwargs)
            if isinstance(ak_arr, ak.Array):
                if where:
                    ak_arr = ak.with_field(arr0, ak_arr, where)
                if frame and ak_arr.fields == ar0[0].fields:
                    return self.to_output(ak_arr).ak.unpack()
                return self.to_output(ak_arr)
            return ak_arr

        return f

    def __init_subclass__(cls, **kwargs):
        # auto add methods to all derivative classes
        cls._add_all()

    @property
    def array(self):
        return ak.from_arrow(self.arrow)

    @property
    def arrow(self):
        return self.to_arrow(self._obj)

    def pack(self):
        """Make a single complex series out of the columns of a dataframe"""
        if not isinstance(self._obj, self.dataframe_type):
            raise ValueError("Can only pack on a dataframe")
        out = {}
        for k in self._obj.columns:
            # TODO: partial pack when column names are like "record.field"
            out[k] = self._obj[k].ak.array
        arr = ak.Array(out)
        return self.to_output(arr)

    def unpack(self):
        """Make dataframe out of a series of record type"""
        # TODO: what to do when passed a dataframe, partial unpack of record fields?
        arr = self.array
        if not arr.fields:
            raise ValueError("Not array-of-records")
        out = {k: self.to_output(arr[k]) for k in arr.fields}
        return self.dataframe_type(out)

    def to_output(self, data=None):
        """Data returned as a series"""
        raise NotImplementedError

    def to_arrow(self, data=None):
        """Data as an arrow array"""
        raise NotImplementedError


class LazyAccessor(EagerAccessor):
    def __getattr__(self, item):
        raise NotImplementedError

    def unexplode(self):
        raise NotImplementedError

    def pack(self):
        raise NotImplementedError

    def unpack(self):
        raise NotImplementedError



from __future__ import annotations

import functools

import awkward as ak
import pyarrow as pa
import pyarrow.compute as pc

from akimbo.apply_tree import dec
from akimbo.mixin import EagerAccessor, LazyAccessor
from akimbo.utils import match_string


def _encode(layout):
    layout._parameters["__array__"] = "bytestring"
    layout.content._parameters["__array__"] = "byte"
    return layout


def match_bytestring(*layout):
    return layout[0].is_list and layout[0].parameter("__array__") == "bytestring"


def _decode(layout):
    layout._parameters["__array__"] = "string"
    layout.content._parameters["__array__"] = "char"
    return layout


_decode_f = dec(_decode, match=match_bytestring, inmode="ak")
_encode_f = dec(_encode, match=match_string, inmode="ak")
_SA_METHODMAPPING = {
    "endswith": "ends_with",
    "isalnum": "is_alnum",
    "isalpha": "is_alpha",
    "isascii": "is_ascii",
    "isdecimal": "is_decimal",
    "isdigit": "is_digit",
    "islower": "is_lower",
    "isnumeric": "is_numeric",
    "isprintable": "is_printable",
    "isspace": "is_space",
    "istitle": "is_title",
    "isupper": "is_upper",
    "startswith": "starts_with",
}
methods = [
    aname
    for aname in (dir(ak.str))
    if not aname.startswith(("_", "akstr_")) and not aname[0].isupper()
]


@functools.wraps(pc.strptime)
def strptime(*args, format="%FT%T", unit="us", error_is_null=True, **kw):
    """strptime with typical defaults set to reverse strftime"""
    out = pc.strptime(
        *args, format=format, unit=unit, error_is_null=error_is_null, **kw
    )
    return out


def repeat(arr, count):
    return pc.binary_repeat(arr, count)


def concat(arr, arr2, sep=""):
    return pc.binary_join_element_wise(
        arr.cast(pa.string()), arr2.cast(pa.string()), sep
    )


class StringAccessor:
    """String operations on nested/var-length data"""

    # TODO: implement dunder add (concat strings) and mul (repeat strings)
    #  - s.ak.str + "suffix" (and arguments swapped)
    #  - s.ak.str + s2.ak.str (with matching schemas)
    #  - s.ak.str * N (and arguments swapped)
    #  - s.ak.str * s (where each string maps to integers for variable repeats)

    def encode(self, arr, encoding: str = "utf-8"):
        """Encode Series of strings to Series of bytes. Leaves non-strings alone."""
        if encoding.lower() not in ["utf-8", "utf8"]:
            raise NotImplementedError
        return _encode_f(arr)

    def decode(self, arr, encoding: str = "utf-8"):
        """Decode Series of bytes to Series of strings. Leaves non-bytestrings alone.

        Validity of UTF8 is *not* checked.
        """
        if encoding.lower() not in ["utf-8", "utf8"]:
            raise NotImplementedError
        return _decode_f(arr)

    @staticmethod
    def method_name(attr: str) -> str:
        return _SA_METHODMAPPING.get(attr, attr)

    def __getattr__(self, attr: str) -> callable:
        attr = self.method_name(attr)
        return getattr(ak.str, attr)

    strptime = staticmethod(dec(strptime, match=match_string, inmode="arrow"))
    repeat = staticmethod(dec(repeat, match=match_string, inmode="arrow"))
    join_el = staticmethod(dec(concat, match=match_string, inmode="arrow"))

    def __add__(self, *_):
        return dec(concat, match=match_string, inmode="arrow")

    def __mul__(self, *_):
        return dec(repeat, match=match_string, inmode="arrow")

    def __dir__(self) -> list[str]:
        return sorted(methods + ["strptime", "encode", "decode"])


EagerAccessor.register_accessor("str", StringAccessor)
LazyAccessor.register_accessor("str", StringAccessor)



from __future__ import annotations

import awkward as ak


class NoAttributes:
    """Allows importing akimbo.cudf even if cudf isn't installed

    This is done so that sphinx can still build docs on non-GPU systems.
    """

    def __dir__(self):
        return []

    def __getattr__(self, item):
        if item == "__qualname__":
            return "akimbo.utils.DummyAttributesObject"
        if item == "__type_params__":
            return ()
        return self

    def __call__(self, *args, **kwargs):
        return self

    __name__ = "DummyAttributesObject"
    __doc__ = None
    __annotations__ = None


def to_ak_layout(ar, backend=None):
    if hasattr(ar, "ak"):
        return ar.ak.array
    elif hasattr(ar, "array"):
        return ar.array
    elif isinstance(ar, (ak.Array)):
        return ar
    else:
        return ak.Array(ak.to_layout(ar))


def match_string(*layout):
    return layout[0].is_list and layout[0].parameter("__array__") == "string"


def rec_list_swap(arr: ak.Array, field: str | None = None) -> ak.Array:
    """Make a record-of-lists into a list-of-records, assuming the lists have the same lengths"""
    record_of_lists = arr[field] if field else arr
    list_of_records = ak.zip(
        dict(zip(ak.fields(record_of_lists), ak.unzip(record_of_lists))), depth_limit=2
    )
    return ak.with_field(arr, list_of_records, field) if field else list_of_records



from typing import Dict

import awkward as ak
import polars as pl
import pyarrow as pa

from akimbo.mixin import EagerAccessor, LazyAccessor


@pl.api.register_series_namespace("ak")
@pl.api.register_dataframe_namespace("ak")
class PolarsAwkwardAccessor(EagerAccessor):
    """Perform awkward operations on a polars series or dataframe

    This is for *eager* operations. A Lazy version may eventually be made.
    """

    series_type = pl.Series
    dataframe_type = pl.DataFrame

    @classmethod
    def to_arrow(cls, data):
        return data.to_arrow()

    def pack(self):
        # polars already implements this directly
        return self._obj.to_struct()

    def to_output(self, data=None):
        arr = data if data is not None else self._obj
        pa_arr = ak.to_arrow(arr, extensionarray=False)
        return pl.from_arrow(pa_arr)


@pl.api.register_lazyframe_namespace("ak")
class LazyPolarsAwkwardAccessor(LazyAccessor):
    dataframe_type = pl.LazyFrame
    series_type = None  # lazy is never series

    def to_output(self, data=None):
        out = self._obj.collect()
        if out.columns == ["_ak_series_"]:
            out = out["_ak_series_"]
        return out

    def __getattr__(self, item: str, **flags) -> callable:
        if isinstance(item, str) and item in self.subaccessors:
            return LazyPolarsAwkwardAccessor(
                self._obj, subaccessor=item, behavior=self._behavior
            )

        def select(*inargs, subaccessor=self.subaccessor, where=None, **kwargs):
            if subaccessor and isinstance(item, str):
                func0 = getattr(self.subaccessors[subaccessor](), item)
            elif callable(item):
                func0 = item
            else:
                func0 = None

            def f(batch):
                arr = ak.from_arrow(batch.to_arrow())

                if any(isinstance(_, str) and _ == "_ak_other_" for _ in inargs):
                    # binary input
                    other = arr[[_ for _ in arr.fields if _.startswith("_df2_")]]
                    # 5 == len("_df2_"); rename to original fields
                    other.layout._fields[:] = [k[5:] for k in other.fields]
                    arr = arr[[_ for _ in arr.fields if not _.startswith("_df2_")]]
                    if other.fields == ["_ak_series_"]:
                        other = other["_ak_series_"]
                    if where is not None:
                        other = other[where]
                    inargs0 = [other if str(_) == "_ak_other_" else _ for _ in inargs]
                else:
                    inargs0 = inargs
                if where:
                    arr0 = arr
                    arr = arr[where]
                if arr.fields == ["_ak_series_"]:
                    arr = arr["_ak_series_"]

                out = func0(arr, *inargs0, **kwargs)
                if where:
                    out = ak.with_field(arr0, out, where)
                if not out.layout.fields:
                    out = ak.Array({"_ak_series_": out})
                arr = ak.to_arrow_table(out, extensionarray=False)
                return pl.DataFrame(arr, **flags)

            inargs = [_._obj if isinstance(_, type(self)) else _ for _ in inargs]
            n_others = sum(isinstance(_, self.dataframe_type) for _ in inargs)
            if n_others == 1:
                other = next(_ for _ in inargs if isinstance(_, self.dataframe_type))
                inargs = [
                    "_ak_other_" if isinstance(_, self.dataframe_type) else _
                    for _ in inargs
                ]
                obj = concat(self._obj, other)
            elif n_others > 1:
                raise NotImplementedError
            else:
                obj = self._obj
            arrow_type = polars_to_arrow_schema(obj.collect_schema())
            arr = pa.table([[]] * len(arrow_type), schema=arrow_type)
            out1 = f(pl.from_arrow(arr))

            return obj.map_batches(f, schema=out1.schema)

        return select

    def pack(self):
        return self._obj.select(
            pl.struct(*self._obj.collect_schema().names()).alias("_ak_series_")
        )

    def unpack(self):
        cols = self._obj.collect_schema().names()
        assert len(cols) == 1
        return self._obj.select(pl.col(cols[0]).struct.unnest())


def concat(*series: pl.LazyFrame) -> pl.LazyFrame:
    this, *others = series
    # don't actually expect more than one "others"
    return this.with_columns(
        [
            o.rename({c: f"_df{i + 2}_{c}" for c in o.collect_schema().names()})
            for i, o in enumerate(others)
        ]
    )


def polars_to_arrow_type(polars_type: pl.DataType) -> pa.DataType:
    type_mapping = {
        pl.Int8: pa.int8(),
        pl.Int16: pa.int16(),
        pl.Int32: pa.int32(),
        pl.Int64: pa.int64(),
        pl.UInt8: pa.uint8(),
        pl.UInt16: pa.uint16(),
        pl.UInt32: pa.uint32(),
        pl.UInt64: pa.uint64(),
        pl.Float32: pa.float32(),
        pl.Float64: pa.float64(),
        pl.String: pa.string(),
        pl.Boolean: pa.bool_(),
        pl.Date: pa.date32(),
    }

    if polars_type in type_mapping:
        return type_mapping[polars_type]

    # parametrised types
    if isinstance(polars_type, pl.Datetime):
        return pa.timestamp(polars_type.unit, polars_type.time_zone)

    if isinstance(polars_type, pl.Decimal):
        return pa.decimal128(polars_type.precision, polars_type.scale)

    # Handle list type
    if isinstance(polars_type, pl.List):
        value_type = polars_to_arrow_type(polars_type.inner)
        return pa.list_(value_type)

    # Handle struct type
    if isinstance(polars_type, pl.Struct):
        fields = []
        for name, dtype in dict(polars_type).items():
            arrow_type = polars_to_arrow_type(dtype)
            fields.append(pa.field(name, arrow_type))
        return pa.struct(fields)

    raise ValueError(f"Unsupported Polars type: {polars_type}")


def polars_to_arrow_schema(polars_schema: Dict[str, pl.DataType]) -> pa.Schema:
    fields = []
    for name, dtype in polars_schema.items():
        arrow_type = polars_to_arrow_type(dtype)
        fields.append(pa.field(name, arrow_type))
    return pa.schema(fields)



from __future__ import annotations

import functools
import inspect
from typing import Callable, Literal

import awkward as ak
import pyarrow as pa


def match_any(*layout, **_):
    return True


def leaf(*layout, **_):
    """True for the lowest elements of any akwward layout tree"""
    return layout[0].is_leaf


def numeric(*layout, **_):
    return layout[0].is_leaf and layout[0].parameters.get("__array__", None) not in {
        "string",
        "char",
    }


def run_with_transform(
    arr: ak.Array,
    op: callable,
    match: callable = leaf,
    outtype: callable | None = None,
    inmode: str = "ak",
    others=(),
    match_kwargs: dict | None = None,
    **kw,
) -> ak.Array:
    def func(layout, **kwargs):
        from akimbo.utils import match_string

        if not isinstance(layout, tuple):
            layout = (layout,)
        if all(match(lay, **(match_kwargs or {})) for lay in layout):
            if inmode == "arrow":
                out = ak.str._apply_through_arrow(
                    op, *layout, **kw, **(match_kwargs or {})
                )
            elif inmode == "numpy":
                # works on numpy/cupy contents
                out = op(*(lay.data for lay in layout), **kw, **(match_kwargs or {}))
            elif inmode == "ak":
                out = op(*layout, **kw, **(match_kwargs or {}))
            else:
                out = op(
                    *(ak.Array(lay) for lay in layout), **kw, **(match_kwargs or {})
                )
            if callable(outtype):
                return outtype(out)
            elif isinstance(out, ak.Array):
                return out.layout
            else:
                return out
        if match_string(*layout):
            # non-string op may fail to descend into string
            return layout[0]

    return ak.transform(func, arr, *others, allow_records=True)


def dec(
    func: callable,
    match: Callable[[ak.contents.Content], bool] = leaf,
    outtype: Callable[[ak.contents.Content], ak.contents.Content] | None = None,
    inmode: Literal["arrow", "numpy", "ak", "other"] = "ak",
    match_kwargs=None,
):
    """Make a nested/ragged version of an operation to apply throughout a tree"""
    match_kw = match_kwargs

    @functools.wraps(func)
    def f(arr, *args, where=None, match_kwargs=None, **kwargs):
        others = []
        if args:
            sig = list(inspect.signature(func).parameters)[1:]
            for k, arg in zip(sig, args):
                if isinstance(arg, ak.Array):
                    others.append(arg)
                elif hasattr(arg, "ak") and hasattr(arg.ak, "array"):
                    others.append(arg.ak.array)
                elif isinstance(arg, pa.Array):
                    others.append(ak.from_arrow(arg))
                else:
                    kwargs.update({k: arg for k, arg in zip(sig, args)})
        if where:
            bits = tuple(where.split(".") if isinstance(where, str) else where)
            part = arr.__getitem__(bits)
            others = [
                o.__getitem__(bits) if isinstance(o, ak.Array) else o for o in others
            ]
        else:
            part = arr

        out = run_with_transform(
            part,
            func,
            match=match,
            outtype=outtype,
            inmode=inmode,
            others=others,
            match_kwargs=match_kwargs or match_kw,
            **kwargs,
        )
        if where:
            out = ak.with_field(arr, out, where=where)
        return out

    f.__doc__ = f"""Run vectorized functions on nested/ragged/complex array

where: None | str | Sequence[str, ...]
    if None, will attempt to apply the kernel throughout the nested structure,
    wherever correct types are encountered. If where is given, only the selected
    part of the structure will be considered, but the output will retain
    the original shape. A fieldname or sequence of fieldnames to descend into
    the tree are acceptable
match_kwargs: None | dict
    any extra field identifiers for matching a record as OK to process

{'--Kernel documentation follows from the original function--' if f.__doc__ else ''}

{f.__doc__ or str(f)}
"""

    return f



import awkward as ak
import pyarrow as pa
import pyspark
from pyspark.sql.pandas.types import from_arrow_schema, to_arrow_schema

from akimbo.mixin import LazyAccessor

sdf = pyspark.sql.DataFrame


class SparkAccessor(LazyAccessor):
    """Operations on pyspark dataframes.

    This is a lazy backend, and operates partition-wise. It predicts the schema
    of each operation by running with an empty dataframe of the correct type.
    """

    dataframe_type = sdf

    def to_arrow(self, data) -> pa.Table:
        # collects data locally
        batches = data._collect_as_arrow()
        return pa.Table.from_batches(batches)

    def to_output(self, data=None):
        from akimbo.pandas import pd

        # data is always arrow format internally
        data = self.to_arrow(data if data is not None else self._obj).to_pandas(
            types_mapper=pd.ArrowDtype
        )
        if list(data.columns) == ["_ak_series_"]:
            data = data["_ak_series_"]
        return data

    def __getattr__(self, item: str) -> sdf:
        if isinstance(item, str) and item in self.subaccessors:
            return SparkAccessor(self._obj, subaccessor=item, behavior=self._behavior)

        def select(*inargs, subaccessor=self.subaccessor, where=None, **kwargs):
            if subaccessor and isinstance(item, str):
                func0 = getattr(self.subaccessors[subaccessor](), item)
            elif callable(item):
                func0 = item
            else:
                func0 = None

            def f(batches):
                for batch in batches:
                    arr = ak.from_arrow(batch)
                    if any(isinstance(_, str) and _ == "_ak_other_" for _ in inargs):
                        # binary input
                        arr, other = arr["_1"], arr["_df2"]
                        if other.fields == ["_ak_series_"]:
                            other = other["_ak_series_"]
                        if where is not None:
                            other = other[where]
                        inargs0 = [
                            other if str(_) == "_ak_other_" else _ for _ in inargs
                        ]
                    else:
                        inargs0 = inargs
                        other = None
                    if where:
                        arr0 = arr
                        arr = arr[where]
                    if arr.fields == ["_ak_series_"]:
                        arr = arr["_ak_series_"]

                    if callable(func0):
                        func = func0
                        args = (arr,)
                    elif hasattr(arr, item) and callable(getattr(arr, item)):
                        func = getattr(arr, item)
                        args = ()
                    elif subaccessor:
                        func = func0
                        args = (arr,)
                    elif hasattr(ak, item):
                        func = getattr(ak, item)
                        args = (arr,)
                    else:
                        raise KeyError(item)

                    out = func(*args, *inargs0, **kwargs)
                    if where:
                        out = ak.with_field(arr0, out, where)
                    if not out.layout.fields:
                        out = ak.Array({"_ak_series_": out})
                    arrout = ak.to_arrow(
                        out,
                        extensionarray=False,
                        list_to32=True,
                        string_to32=True,
                        bytestring_to32=True,
                    )
                    yield pa.RecordBatch.from_struct_array(arrout)

            f.__name__ = item.__name__ if callable(item) else item

            inargs = [_._obj if isinstance(_, type(self)) else _ for _ in inargs]
            n_others = sum(isinstance(_, self.dataframe_type) for _ in inargs)
            if n_others == 1:
                other = next(_ for _ in inargs if isinstance(_, self.dataframe_type))
                inargs = [
                    "_ak_other_" if isinstance(_, self.dataframe_type) else _
                    for _ in inargs
                ]
                obj = concat_columns_zip_index(self._obj, other)
            elif n_others > 1:
                raise NotImplementedError
            else:
                obj = self._obj
            arrow_type = to_arrow_schema(obj.schema)
            arr = pa.table([[]] * len(arrow_type), schema=arrow_type)
            out1 = next(f([arr]))
            out_schema = pa.table(out1).schema
            return obj.mapInArrow(f, schema=from_arrow_schema(out_schema))

        return select


def concat_columns_zip_index(df1: sdf, df2: sdf) -> sdf:
    """Add two DataFrames' columns into a single DF.

    The is SQL-tricky, but at least it requires no python map/iteration!
    """
    if df1.rdd.getNumPartitions() != df2.rdd.getNumPartitions():
        # warn that this causes shuffle
        pass
    df1_ind = df1.rdd.zipWithIndex().toDF()
    df2_ind = df2.rdd.zipWithIndex().toDF().withColumnRenamed("_1", "_df2")
    return df1_ind.join(df2_ind, "_2", "left").sort("_2").drop("_2")


@property  # type:ignore
def ak_property(self):
    return SparkAccessor(self)


pyspark.sql.DataFrame.ak = ak_property  # spark has no Series



import pyarrow.compute as pc

from akimbo.apply_tree import dec
from akimbo.mixin import EagerAccessor, LazyAccessor


def match(*layouts):
    return layouts[0].is_leaf and layouts[0].dtype.kind == "M"


methods = [
    "ceil_temporal",
    "day",
    "day_of_week",
    "day_of_year",
    "day_time_interval_between",
    "days_between",
    "floor_temporal",
    "hour",
    "hours_between",
    "is_leap_year",
    "iso_calendar",
    "iso_week",
    "iso_year",
    "microsecond",
    "microseconds_between",
    "millisecond",
    "milliseconds_between",
    "minute",
    "minutes_between",
    "month",
    "month_day_nano_interval_between",
    "month_interval_between",
    "nanosecond",
    "nanoseconds_between",
    "quarter",
    "quarters_between",
    "reound_temporal",
    "second",
    "seconds_between",
    "strftime",
    "subsecond",
    "us_week",
    "us_year",
    "week",
    "weeks_between",
    "year",
    "year_month_day",
    "years_between",
]


class DatetimeAccessor:
    def __getattr__(self, item):
        if item in methods:
            func = getattr(pc, item)
        else:
            raise AttributeError

        return dec(func, match=match, inmode="arrow")

    def __dir__(self):
        return methods


EagerAccessor.register_accessor("dt", DatetimeAccessor)
LazyAccessor.register_accessor("dt", DatetimeAccessor)



import functools
from typing import Callable

import awkward as ak
import numpy as np

from akimbo.utils import NoAttributes

try:
    import cudf
    from cudf import DataFrame, Series
    from cudf import _lib as libcudf
    from cudf.core.column.datetime import DatetimeColumn
    from cudf.core.column.string import StringMethods
except ImportError:
    StringMethods = NoAttributes()
    DatetimeColumn = NoAttributes()
    libcudf = NoAttributes()
    DataFrame = Series = NoAttributes()


from akimbo.ak_from_cudf import cudf_to_awkward as from_cudf
from akimbo.apply_tree import dec, leaf
from akimbo.datetimes import DatetimeAccessor
from akimbo.datetimes import match as match_dt
from akimbo.datetimes import methods as dt_methods
from akimbo.mixin import EagerAccessor, numeric
from akimbo.ops import rename
from akimbo.strings import _SA_METHODMAPPING, StringAccessor, match_string, strptime


def dec_cu(op, match=match_string):
    @functools.wraps(op)
    def f(lay, **kwargs):
        # op(column, ...)->column
        col = op(lay._to_cudf(cudf, None, len(lay)), **kwargs)
        if hasattr(cudf.Series, "_from_column"):
            return from_cudf(cudf.Series._from_column(col)).layout
        return from_cudf(cudf.Series(col)).layout

    return dec(func=f, match=match, inmode="ak")


class CudfStringAccessor(StringAccessor):
    """String operations on nested/var-length data"""

    def decode(self, arr, encoding: str = "utf-8"):
        raise NotImplementedError(
            "cudf does not support bytearray type, so we can't automatically identify them"
        )

    def encode(self, arr, encoding: str = "utf-8"):
        raise NotImplementedError("cudf does not support bytearray type")

    @staticmethod
    def method_name(attr: str) -> str:
        return _SA_METHODMAPPING.get(attr, attr)

    def __getattr__(self, attr: str) -> callable:
        attr = self.method_name(attr)

        @functools.wraps(getattr(StringMethods, attr))
        def f(lay, **kwargs):
            col = lay._to_cudf(cudf, None, len(lay))
            if hasattr(cudf.Series, "_from_column"):
                ser = cudf.Series._from_column(col)
            else:
                ser = cudf.Series(col)
            col = getattr(StringMethods(ser), attr)(**kwargs)
            return from_cudf(col).layout

        return dec(f, match=match_string, inmode="ak")

    strptime = staticmethod(dec_cu(strptime, match=match_string))


class CudfDatetimeAccessor(DatetimeAccessor):
    def __getattr__(self, attr):
        @functools.wraps(getattr(DatetimeColumn, attr))
        def f(lay, **kwargs):
            col = lay._to_cudf(cudf, None, len(lay))
            col = getattr(col, attr)
            if callable(col):
                # as opposed to an attribute
                col = col(**kwargs)
            if hasattr(cudf.Series, "_from_column"):
                ser = cudf.Series._from_column(col)
            else:
                ser = cudf.Series(col)
            return from_cudf(ser).layout

        return dec(f, match=match_dt, inmode="ak")

    def __dir__(self):
        return dt_methods


def _cast_inner(col, dtype):
    try:
        return libcudf.unary.cast(col, dtype)
    except AttributeError:
        return cudf.core.column.ColumnBase(
            col.data,
            size=len(col),
            dtype=np.dtype(dtype),
            mask=None,
            offset=0,
            children=(),
        )


class CudfAwkwardAccessor(EagerAccessor):
    """Operations on cuDF dataframes on the GPU.

    Data are kept in GPU memory and use views rather than copies where
    possible.
    """

    funcs = {"cast": dec_cu(_cast_inner, match=leaf), "rename": rename}

    series_type = Series
    dataframe_type = DataFrame
    subaccessors = {"str": CudfStringAccessor, "dt": CudfDatetimeAccessor}

    def to_output(self, data=None):
        arr = data if data is not None else self._obj
        if isinstance(arr, ak.Array):
            return ak.to_cudf(arr)
        elif isinstance(arr, ak.contents.Content):
            return arr._to_cudf(cudf, None, len(arr))
        return arr

    @property
    def array(self):
        data = self._obj
        if isinstance(data, self.series_type):
            return from_cudf(data)
        out = {}
        for col in data.columns:
            out[col] = from_cudf(data[col])
        return ak.Array(out)

    @classmethod
    def _create_op(cls, op):
        def run(self, *args, **kwargs):
            args = [
                ak.Array([_], backend="cuda")
                if isinstance(_, (str, int, float, np.number))
                else _
                for _ in args
            ]
            return self.transform(op, *args, match=numeric, **kwargs)

        return run

    def apply(self, fn: Callable, *args, **kwargs):
        if "CPUDispatcher" in str(fn):
            # auto wrap original function for GPU numba func
            raise NotImplementedError
        super().apply(fn, *args, **kwargs)


@property  # type:ignore
def ak_property(self):
    return CudfAwkwardAccessor(self)


Series.ak = ak_property  # no official register function?
DataFrame.ak = ak_property  # no official register function?



import awkward as ak
import dask.dataframe as dd
import pandas as pd
import pyarrow as pa
from dask.dataframe.accessor import (
    register_dataframe_accessor,
    register_series_accessor,
)

from akimbo.mixin import LazyAccessor
from akimbo.pandas import PandasAwkwardAccessor


class DaskAwkwardAccessor(LazyAccessor):
    """Perform awkward operations on a dask series or frame

    These operations are lazy, because of how dask works. Note
    that we use mapping operations here, so any action on
    axis==0 or 1 will produce results per partition, which
    you must then combine.

    To perform intra-partition operations, we recommend you
    use the ``.to_dask_awkward`` method.

    Correct arrow dtypes will be deduced when the input is
    also arrow, which is now the default for the dask
    "dataframe.dtype_backend" config options.
    """

    series_type = dd.Series
    dataframe_type = dd.DataFrame

    @staticmethod
    def _to_tt(data):
        # self._obj._meta.convert_dtypes(dtype_backend="pyarrow")
        data = data._meta if hasattr(data, "_meta") else data
        arr = PandasAwkwardAccessor.to_arrow(data)
        if isinstance(arr, pa.ChunkedArray) and len(arr) == 0:
            arr = arr.combine_chunks()
        return ak.to_backend(ak.from_arrow(arr), "typetracer")

    def to_dask_awkward(self):
        """Convert to dask-awkard.Array object

        This make a single complex awkward array type out of one or more columns.
        You would do this, in order to use dask-awkward's more advanced inter
        partition aggregations and column/buffer IO optimisation.
        See https://dask-awkward.readthedocs.io/

        c.f., dask_awkward.to_dataframe
        """
        import dask_awkward as dak

        tt = self._to_tt(self._obj)
        return dak.lib.core.new_array_object(
            self._obj.dask, divisions=self._obj.divisions, name=self._obj._name, meta=tt
        )

    def __getattr__(self, item):
        if self.subaccessor and isinstance(item, str):
            item = getattr(self.subaccessors[self.subaccessor], item)
        elif isinstance(item, str) and item in self.subaccessors:
            return DaskAwkwardAccessor(
                self._obj, subaccessor=item, behavior=self._behavior
            )

        def select(*inargs, where=None, **kwargs):
            orig = self._obj.head()

            def func2(data):
                import akimbo.pandas  # noqa: F401

                others = [
                    (k._obj if isinstance(k, DaskAwkwardAccessor) else k)
                    for k in inargs
                ]
                if isinstance(item, str):
                    # work on pandas API
                    return getattr(PandasAwkwardAccessor(data), item)(*others, **kwargs)
                else:
                    # ak to ak
                    arr = data.ak.array
                    # others =
                    if where:
                        part = arr[where]
                        arr = ak.with_field(arr, part, where)
                        others = [_[where] for _ in others]
                    else:
                        part = arr
                    out = item(part, *others, **kwargs)
                    if where:
                        out = ak.with_field(arr, out, where)
                    out = pd.arrays.ArrowExtensionArray(
                        ak.to_arrow(out, extensionarray=False)
                    )
                    return pd.Series(out)

            out0 = func2(orig)
            return self._obj.map_partitions(func2, meta=out0)

        return select

    def __dir__(self) -> list[str]:
        return sorted(super().__dir__() + ["to_dask_awkward"])


register_series_accessor("ak")(DaskAwkwardAccessor)
register_dataframe_accessor("ak")(DaskAwkwardAccessor)



import awkward as ak


def rename(arr: ak.Array, where: str, to: str) -> ak.Array:
    """Assign new field name to given location in the structure

    Parameters
    ----------
    arr: array to change
    where: str | tuple[str]
        location we would like to rename
    to: str
        new name
    """
    lay = ak.copy(arr.layout)
    where = list(where) if isinstance(where, tuple) else [where]
    parent = None
    bit = lay
    this = None
    while where:
        if getattr(bit, "contents", None):
            this = bit.fields.index(where.pop(0))
            parent, bit = bit, bit.contents[this]
        else:
            parent, bit = bit, bit.content
    parent.fields[this] = to
    return ak.Array(lay)


def join(
    arr: ak.Array,
    other: ak.Array,
    key: str,
    colname: str = "match",
    sort: bool = False,
    rkey: str | None = None,
    numba: bool = True,
):
    """DB ORM-style left join to other dataframe/series with nesting but no copy

    Related records of the ``other`` table will appear as a list under the new field
    ``colname`` for all matching keys. This is the speed and memory efficient way
    to doing a pandas-style merge/join, which explodes out the values to a much
    bigger memory footprint.

    Parameters
    ----------
    other: series or table
    key: name of the field in this table to match on
    colname: the field that will be added to each record. This field will exist even
        if there are no matches, in which case the list will be empty.
    sort: if False, assumes that they key is sorted in both tables. If True, an
        argsort is performed first, and the match is done by indexing. This may be
        significantly slower.
    rkey: if the name of the field to match on is different in the ``other`` table.
    numba: the matching algorithm will go much faster using numba. However, you can
        set this to False if you do not have numba installed.
    """
    from akimbo.io import join

    out = join(arr, other, key, colname=colname, sort=sort, rkey=rkey, numba=numba)
    return out



from collections.abc import Iterable

import awkward as ak
import pandas as pd
import pyarrow
import pyarrow as pa

from akimbo.mixin import EagerAccessor


@pd.api.extensions.register_series_accessor("ak")
@pd.api.extensions.register_dataframe_accessor("ak")
class PandasAwkwardAccessor(EagerAccessor):
    """Perform awkward operations on pandas data

    Nested structures are handled using arrow as the
    storage backend. If you use pandas object columns
    (python lists, dicts, strings), they will be converted
    on any access to a ``.ak`` method.
    """

    series_type = pd.Series
    dataframe_type = pd.DataFrame

    @classmethod
    def to_arrow(cls, data):
        if isinstance(data, ak.Array):
            return ak.to_arrow(data)
        if isinstance(data, ak.Record):
            return ak.to_arrow_table(data)
        if isinstance(data, (pyarrow.Array, pyarrow.Table)):
            return data
        if isinstance(data, cls.series_type):
            if getattr(data.dtype, "storage", "") == "pyarrow":
                return data.array.__arrow_array__()
            return pa.array(data)
        return pa.table(data)

    def to_output(self, data=None):
        # override to apply index
        data: ak.Array = data if data is not None else self.array
        if not isinstance(data, Iterable):
            return data
        arr = pd.arrays.ArrowExtensionArray(ak.to_arrow(data, extensionarray=False))
        if self._obj is not None and len(arr) == len(self._obj.index):
            return pd.Series(arr, index=self._obj.index)
        else:
            return pd.Series(arr)

    @staticmethod
    def _validate(_):
        # required by pandas
        return True



import awkward as ak
import pyarrow as pa
import ray
import ray.data as rd

from akimbo.mixin import LazyAccessor


class RayAccessor(LazyAccessor):
    """Operations on ray.data.Dataset dataframes.

    This is a lazy backend, and operates partition-wise. It predicts the schema
    of each operation by running with an empty dataframe of the correct type.
    """

    dataframe_type = rd.Dataset
    series_type = None  # only has "dataframe like"

    def to_arrow(self, data: rd.Dataset) -> pa.Table:
        batches = ray.get(data.to_arrow_refs())
        return pa.concat_tables(batches)

    def to_output(self, data=None):
        import pandas as pd

        data = self.to_arrow(data if data is not None else self._obj)
        data = data.to_pandas(types_mapper=pd.ArrowDtype)
        if list(data.columns) == ["_ak_series_"]:
            data = data["_ak_series_"]
        return data

    def __getattr__(self, item: str) -> callable:
        if isinstance(item, str) and item in self.subaccessors:
            return RayAccessor(self._obj, subaccessor=item, behavior=self._behavior)

        def select(*inargs, subaccessor=self.subaccessor, where=None, **kwargs):
            if subaccessor and isinstance(item, str):
                func0 = getattr(self.subaccessors[subaccessor](), item)
            elif callable(item):
                func0 = item
            else:
                func0 = None

            def f(batch):
                arr = ak.from_arrow(batch)
                if any(isinstance(_, str) and _ == "_ak_other_" for _ in inargs):
                    # binary input
                    other = arr[[_ for _ in arr.fields if _.startswith("_df2_")]]
                    # 5 == len("_df2_"); rename to original fields
                    other.layout._fields[:] = [k[5:] for k in other.fields]
                    arr = arr[[_ for _ in arr.fields if not _.startswith("_df2_")]]
                    if other.fields == ["_ak_series_"]:
                        other = other["_ak_series_"]
                    if where is not None:
                        other = other[where]
                    inargs0 = [other if str(_) == "_ak_other_" else _ for _ in inargs]
                else:
                    inargs0 = inargs
                if where:
                    arr0 = arr
                    arr = arr[where]
                if arr.fields == ["_ak_series_"]:
                    arr = arr["_ak_series_"]

                if callable(func0):
                    func = func0
                    args = (arr,)
                elif hasattr(arr, item) and callable(getattr(arr, item)):
                    func = getattr(arr, item)
                    args = ()
                elif subaccessor:
                    func = func0
                    args = (arr,)
                elif hasattr(ak, item):
                    func = getattr(ak, item)
                    args = (arr,)
                else:
                    raise KeyError(item)

                out = func(*args, *inargs0, **kwargs)
                if where:
                    out = ak.with_field(arr0, out, where)
                if not out.layout.fields:
                    out = ak.Array({"_ak_series_": out})
                return ak.to_arrow_table(
                    out,
                    extensionarray=False,
                    list_to32=True,
                    string_to32=True,
                    bytestring_to32=True,
                )

            f.__name__ = item.__name__ if callable(item) else item

            inargs = [_._obj if isinstance(_, type(self)) else _ for _ in inargs]
            n_others = sum(isinstance(_, self.dataframe_type) for _ in inargs)
            if n_others == 1:
                other = next(_ for _ in inargs if isinstance(_, self.dataframe_type))
                inargs = [
                    "_ak_other_" if isinstance(_, self.dataframe_type) else _
                    for _ in inargs
                ]
                obj = concat_columns_zip_index(self._obj, other)
            elif n_others > 1:
                raise NotImplementedError
            else:
                obj = self._obj
            arrow_type = obj.schema().base_schema
            arr = pa.table([[]] * len(arrow_type), schema=arrow_type)
            out1 = f(arr)
            out_schema = pa.table(out1).schema
            result = obj.map_batches(f, zero_copy_batch=True, batch_format="pyarrow")
            result._plan.cache_schema(out_schema)
            return result

        return select


def concat_columns_zip_index(df1: rd.Dataset, df2: rd.Dataset) -> rd.Dataset:
    """Add two DataFrames' columns into a single DF."""
    if df1.num_blocks != df2.num_blocks:
        # warn that this causes shuffle
        pass
    return df1.zip(df2.rename_columns({k: f"_df2_{k}" for k in df2.columns()}))


@property  # type:ignore
def ak_property(self):
    return RayAccessor(self)


rd.Dataset.ak = ak_property  # Ray has no Series



import uuid

import awkward as ak
import duckdb
import duckdb.duckdb.typing as dtyp
import pyarrow as pa

from akimbo.mixin import LazyAccessor


class DuckAccessor(LazyAccessor):
    dataframe_type = duckdb.DuckDBPyRelation
    series_type = None  # only has "dataframe like"

    @classmethod
    def to_arrow(cls, data: duckdb.DuckDBPyRelation):
        return data.to_arrow_table()

    def to_output(self, data=None):
        import pandas as pd

        data = self.to_arrow(data or self._obj)
        if data.column_names == ["_ak_series_"]:
            data = data["_ak_series_"].to_pandas(types_mapper=pd.ArrowDtype)
        elif data.column_names == ["_ak_dataframe_"]:
            data = (
                data["_ak_dataframe_"].to_pandas(types_mapper=pd.ArrowDtype).ak.unpack()
            )
        return data

    def __getattr__(self, item: str) -> callable:
        if isinstance(item, str) and item in self.subaccessors:
            return DuckAccessor(self._obj, subaccessor=item, behavior=self._behavior)

        def select(
            *inargs, subaccessor=self.subaccessor, where=None, **kwargs
        ) -> duckdb.DuckDBPyRelation:
            if subaccessor:
                func0 = getattr(self.subaccessors[subaccessor](), item)
            elif callable(item):
                func0 = item
            else:
                func0 = None

            def f(batch: pa.ChunkedArray) -> pa.Table:
                # length 0 indicates this is a test run, and we want to know the output type: series/dataframe
                ret_type = len(batch) == 0
                # fixes bug: from_arrow fails on ChunkedArray with zero length
                arr = ak.from_arrow(batch if len(batch) else batch.chunks[0])
                arr0 = arr

                if any(isinstance(_, str) and _ == "_ak_other_" for _ in inargs):
                    # binary input
                    other = arr[[_ for _ in arr.fields if _.startswith("_df2_")]]
                    # 5 == len("_df2_"); rename to original fields
                    other.fields[:] = [k[5:] for k in other.fields]
                    arr = arr[[_ for _ in arr.fields if not _.startswith("_df2_")]]
                    if other.fields == ["_ak_series_"]:
                        other = other["_ak_series_"]
                    if where is not None:
                        other = other[where]
                    inargs0 = [other if str(_) == "_ak_other_" else _ for _ in inargs]
                else:
                    inargs0 = inargs
                if arr.fields == ["_ak_series_"]:
                    arr = arr["_ak_series_"]
                elif arr.layout.is_option:
                    # A whole row cannot be NULL unless is a series
                    arr = ak.Array(arr.layout.content)

                if where:
                    arr = arr[where]

                if callable(func0):
                    func = func0
                    args = (arr,)
                elif hasattr(arr, item) and callable(getattr(arr, item)):
                    func = getattr(arr, item)
                    args = ()
                elif hasattr(ak, item):
                    func = getattr(ak, item)
                    args = (arr,)
                else:
                    raise KeyError(item)

                out = func(*args, *inargs0, **kwargs)
                if where:
                    out = ak.with_field(arr0, out, where)
                outtype = "dataframe" if out.layout.is_record else "series"
                if len(out.fields) != 1:
                    out = ak.Array({"_ak_series_": out})
                out = ak.to_arrow_table(
                    out,
                    extensionarray=False,
                    list_to32=True,
                    string_to32=True,
                    bytestring_to32=True,
                )
                # must return one-column table
                if ret_type:
                    return out, outtype
                return out

            f.__name__ = item.__name__ if callable(item) else item

            inargs = [_._obj if isinstance(_, type(self)) else _ for _ in inargs]
            n_others = sum(isinstance(_, self.dataframe_type) for _ in inargs)
            if n_others == 1:
                other = next(_ for _ in inargs if isinstance(_, self.dataframe_type))
                inargs = [
                    "_ak_other_" if isinstance(_, self.dataframe_type) else _
                    for _ in inargs
                ]
                obj = pack(positional_join(self._obj, other))

            elif n_others > 1:
                raise NotImplementedError
            else:
                obj = self._obj

            if len(obj.dtypes) > 1:
                obj = (
                    self.pack()
                )  # no-compute operation, but input has to be a single column

            arrow_type = pa.schema(
                [(obj.columns[0], duckdb_to_pyarrow_type(obj.dtypes[0]))]
            )
            arr = pa.table([[]], schema=arrow_type)
            out1, outtype = f(arr[obj.columns[0]])  # ; used implicitly in next line

            out2 = duckdb.sql("SELECT * FROM out1").dtypes[0]
            fname = f"{item.__name__ if hasattr(item, '__name__') else item}_{uuid.uuid4().hex}"
            duckdb.create_function(
                fname, f, obj.types, out2, type=duckdb.functional.PythonUDFType.ARROW
            )
            # this operates on the one known column, so in the function the column
            # has been normalised out
            result = duckdb.sql(
                f"SELECT {fname}({obj.columns[0]}) as _ak_{outtype}_ FROM obj"
            )
            return result

        return select

    def pack(self):
        return pack(self._obj)

    def unpack(self):
        return unpack(self._obj)


def pack(obj, name="_ak_series_"):
    inner = ",".join(f"{k} := {k}" for k in obj.columns)
    return duckdb.sql(f"SELECT struct_pack({inner}) as {name} FROM obj")


def unpack(obj):
    assert [_.id for _ in obj.dtypes] == ["struct"]
    columns = obj.dtypes[0].children
    inner = ", ".join(
        f"struct_extract({obj.columns[0]}, '{c}') as {c}" for c in columns
    )
    return duckdb.sql(f"SELECT {inner} FROM obj")


def positional_join(
    df1: duckdb.DuckDBPyRelation, df2: duckdb.DuckDBPyRelation
) -> duckdb.DuckDBPyRelation:
    # https://duckdb.org/docs/sql/query_syntax/from#positional-joins
    inner = ", ".join(
        [f"df1.{c} AS {c}" for c in df2.columns]
        + [f"df2.{c} AS _df2_{c}" for c in df2.columns]
    )
    return duckdb.sql(f"SELECT {inner} FROM df1 POSITIONAL JOIN df2")


def _pack_struct_type(obj):
    return duckdb.struct_type({c: t for c, t in zip(obj.columns, obj.dtypes)})


def _unpack_struct_type(obj):
    return dict(obj.dtypes[0].children)


def duckdb_to_pyarrow_type(duckdb_type: dtyp.DuckDBPyType) -> pa.DataType:
    """Convert complex schema"""
    type_map = {
        "tinyint": pa.int8(),
        "smallint": pa.int16(),
        "integer": pa.int32(),
        "bigint": pa.int64(),
        "utinyint": pa.uint8(),
        "usmallint": pa.uint16(),
        "uinteger": pa.uint32(),
        "ubigint": pa.uint64(),
        "float": pa.float32(),
        "real": pa.float32(),
        "double": pa.float64(),
        "boolean": pa.bool_(),
        "varchar": pa.string(),
        "text": pa.string(),
        "date": pa.date32(),
        "time": pa.time64("us"),
        # timez ?
        "timestamp": pa.timestamp("us"),
        "timestamp_ns": pa.timestamp("ns"),
        "timestam_ms": pa.timestamp("ms"),
        "timestamp_s": pa.timestamp("s"),
        # timestampz ?
        "blob": pa.binary(),  # same
    }

    if duckdb_type.id in type_map:
        return type_map[duckdb_type.id]

    # Handle DECIMAL types
    if duckdb_type.id == "decimal":
        return pa.decimal128(**dict(duckdb_type.children))

    # Handle array types
    if duckdb_type.id == "list":
        inner_type = duckdb_to_pyarrow_type(duckdb_type.children[0][1])
        return pa.list_(inner_type)

    # Handle struct types
    if duckdb_type.id == "struct":
        fields = {k: duckdb_to_pyarrow_type(v) for k, v in duckdb_type.children}
        return pa.struct(fields)

    # fixed-length-tuple?
    raise ValueError(f"Unsupported DuckDB type: {duckdb_type}")


@property  # type:ignore
def ak_property(self):
    return DuckAccessor(self)


duckdb.DuckDBPyRelation.ak = ak_property  # Duck has no Series
